@@ -49,17 +49,4 @@ impl<C, E> ReadableCache<E> for C
     {
         Cache::get(self, cacheable).await
     }
-}
-
-pub struct NopCache {}
-
-
-#[async_trait]
-impl Cache<()> for NopCache {
-    async fn get<CA: Cacheable + Send + Sync + 'static>(&mut self, _cacheable: CA) -> Result<CacheResponse, ()> {
-        Ok(CacheResponse::Miss)
-    }
-    async fn store(&mut self, _identity: Vec<u8>) -> Result<(), ()> {
-        Ok(())
-    }
 }
\ No newline at end of file
